#include "buttontimer.hpp"

#include <time.h>
#ifndef CLOCK_MONOTIC
#   define CLOCK_MONOTIC CLOCK_REALTIME
#endif

ButtonTimer::ButtonTimer(uint32_t shortpress_min_ms, uint32_t longpress_ms)
{
    shortpressMinTime = shortpress_min_ms;
    longpressTime = longpress_ms;
    eventLock = false;
    
    ThreadStart();
}

ButtonTimer::~ButtonTimer()
{
    ThreadStop();

}

void ButtonTimer::RegisterPress(const uint8_t keycode)
{
    MutexLock();
    // Prevent trouble when calling this from within one of our event listeners
    if(eventLock) { MutexUnlock(); return; }

    // register current registered time for first triggering press
    if(!pressRegistry.count(keycode))
    {
        pressRegistry[keycode] = now_ms();
    }

    // register press state to true for debouncing
    pressState[keycode] = true;
    MutexUnlock();
}


void ButtonTimer::RegisterRelease(const uint8_t keycode)
{
    MutexLock();
    // Prevent trouble when calling this from within one of our event listeners
    if(eventLock) { MutexUnlock(); return; }

    // register current press state to false
    pressState[keycode] = false;

    if(pressRegistry.count(keycode))
    {
        // if it was a long press, the key code would already have been erased, so we
        // can safely fire the onShortPress event
        uint64_t now = now_ms();
        uint64_t then = pressRegistry[keycode];
        
        if(pressHold.count(keycode))
        {
            // debounce locks for release
            // pressHold[keycode] == true means wait for release
            // pressHold[keycode] == false means we're in the release debounce time
            if(pressHold[keycode] == true){
                // first release after longpress
                // Register time and switch to next mode
                pressHold[keycode] = false;
                pressRegistry[keycode] = now;
            }
            else
            {
                // still in debounce
                // release the debounce lock if this release occurs after the shortpresstime
                if(now - then > shortpressMinTime){
                    pressHold.erase(keycode); 
                    pressRegistry.erase(keycode); 
                }
            }
        }
        else if(now - then > shortpressMinTime)
        {
            // only trigger this if the release is later than the shortpresstime
            // remove from registry after release, if it was a long press, the event should have already been fired
            pressRegistry.erase(keycode); 

            eventLock = true; // lock out trouble
            onShortPress(keycode);
            eventLock = false; // risk of trouble gone
        }
    }

    MutexUnlock();
}

void ButtonTimer::CancelPress(const uint8_t keycode)
{
    // remove button id from map (but only if it is in the map already)
    MutexLock();
    // Prevent trouble when calling this from within one of our event listeners
    if(eventLock) { MutexUnlock(); return; }

    if(pressRegistry.count(keycode))
    {
        pressRegistry.erase(keycode);
    }
    MutexUnlock();
}

void ButtonTimer::ThreadLoop()
{
    uint64_t now = now_ms();
    std::list<uint8_t> shortPressList;
    std::list<uint8_t> longPressList;
    boost::optional<bool> valid;
    MutexLock();
    std::list<uint8_t> removeList;
    // list through all the items 

    //FIXED: Do not alter press registry while iterating through it - that causes segfaults
    for( auto const& itm : pressRegistry)
    {
        if(now - (itm.second) >= shortpressMinTime && pressState[itm.first] == false) // if the shortpresstime wa exceeded and the value is false
        {
            if(!pressHold.count(itm.first))
            {
                shortPressList.push_back(itm.first);
            }
            else if(pressHold[itm.first] == false)
            {
                removeList.push_back(itm.first);
                // release debounce lock from longpress
                // pressRegistry.erase(itm.first); will be done later
                pressHold.erase(itm.first);
            }
        }
        if(now - (itm.second) >= longpressTime && !pressHold.count(itm.first)) // if it is in overtime and not waiting for release debounce
        {
            // Schedule to process the thing
            longPressList.push_back(itm.first);
        }

    }
    // remove all items slated for removal from the pressRegistry
    for( auto const& gpio: removeList)
    {
        pressRegistry.erase(gpio);
    }


    for (auto const& gpio : shortPressList)
    {
        pressRegistry.erase(gpio);
        // If any validators are connected, they can retun false to indicate that this connection is not 
        // allowed
        eventLock = true; // lock out trouble
        onShortPress(gpio);
        eventLock = false; // risk of trouble gone
    }

    // Process listed items
    for (auto const& gpio : longPressList)
    {
        pressHold[gpio] = true; // debounce wait state
        // If any validators are connected, they can retun false to indicate that this connection is not 
        // allowed
        eventLock = true; // lock out trouble
        
        valid = onValidatePress(gpio);
        if(valid.get_value_or(true))
        {
            onLongPress(gpio);
        }
        eventLock = false; // risk of trouble gone
    }
    
    MutexUnlock();
    usleep(50000);
}

int64_t ButtonTimer::now_ms(void)
{
    struct timespec now;
    clock_gettime(CLOCK_MONOTIC, &now);

    return ((int64_t)now.tv_sec)*1000LL + (now.tv_nsec/1000000);
}
